# Copyright (C) 2021  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from typing import Any, Dict

import pytest
import yaml

from swh.core.api import RPCServerApp
from swh.counters.api import server
from swh.counters.api.server import load_and_check_config, make_app_from_configfile


def teardown_function():
    # Ensure there is no configuration loaded from a previous test
    server.api = None


@pytest.fixture
def swh_counters_server_config() -> Dict[str, Any]:
    return {"counters": {"cls": "redis", "hosts": "redis",}}


@pytest.fixture
def swh_counters_server_config_on_disk(
    tmp_path, monkeypatch, swh_counters_server_config
) -> str:
    return _environment_config_file(tmp_path, monkeypatch, swh_counters_server_config)


def write_config_file(tmpdir, config_dict: Dict, name: str = "config.yml") -> str:
    """Prepare configuration file in `$tmpdir/name` with content `content`.

    Args:
        tmpdir (LocalPath): root directory
        content: Content of the file either as string or as a dict.
                            If a dict, converts the dict into a yaml string.
        name: configuration filename

    Returns
        path of the configuration file prepared.

    """
    config_path = tmpdir / name
    config_path.write_text(yaml.dump(config_dict), encoding="utf-8")
    # pytest on python3.5 does not support LocalPath manipulation, so
    # convert path to string
    return str(config_path)


def _environment_config_file(tmp_path, monkeypatch, content):
    conf_path = write_config_file(tmp_path, content)
    monkeypatch.setenv("SWH_CONFIG_FILENAME", conf_path)


@pytest.mark.parametrize("config_file", [None, ""])
def test_load_and_check_config_no_configuration(config_file):
    """Inexistent configuration files raises"""
    with pytest.raises(EnvironmentError, match="Configuration file must be defined"):
        load_and_check_config(config_file)


def test_load_and_check_config_inexistent_file():
    config_path = "/some/inexistent/config.yml"
    expected_error = f"Configuration file {config_path} does not exist"
    with pytest.raises(EnvironmentError, match=expected_error):
        load_and_check_config(config_path)


def test_load_and_check_config_wrong_configuration(tmpdir):
    """Wrong configuration raises"""
    config_path = write_config_file(tmpdir, {"something": "useless"})
    with pytest.raises(KeyError, match="Missing 'counters' configuration"):
        load_and_check_config(config_path)


def test_server_make_app_from_config_file(swh_counters_server_config_on_disk):
    app = make_app_from_configfile()

    assert app is not None
    assert isinstance(app, RPCServerApp)

    app2 = make_app_from_configfile()
    assert app is app2


def test_server_index(swh_counters_server_config_on_disk, mocker):
    """Test the result of the main page"""

    app = make_app_from_configfile()
    app.config["TESTING"] = True
    tc = app.test_client()

    r = tc.get("/")

    assert 200 == r.status_code
    assert b"SWH Counters" in r.get_data()



# Copyright (C) 2021  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import logging
from typing import Any, Iterable

from redis.client import Redis as RedisClient
from redis.exceptions import ConnectionError

DEFAULT_REDIS_PORT = 6379


logger = logging.getLogger(__name__)


class Redis:
    """Redis based implementation of the counters.
       It uses one HyperLogLog collection per counter"""

    _redis_client = None

    def __init__(self, host: str):
        host_port = host.split(":")

        if len(host_port) > 2:
            raise ValueError("Invalid server url `%s`" % host)

        self.host = host_port[0]
        self.port = int(host_port[1]) if len(host_port) > 1 else DEFAULT_REDIS_PORT

    @property
    def redis_client(self) -> RedisClient:

        if self._redis_client is None:
            self._redis_client = RedisClient(host=self.host, port=self.port)

        return self._redis_client

    def check(self):
        try:
            return self.redis_client.ping()
        except ConnectionError:
            logger.exception("Unable to connect to the redis server")
            return False

    def add(self, collection: str, keys: Iterable[Any]) -> None:
        redis = self.redis_client
        pipeline = redis.pipeline(transaction=False)

        [pipeline.pfadd(collection, key) for key in keys]

        pipeline.execute()

    def get_count(self, collection: str) -> int:
        return self.redis_client.pfcount(collection)



# Copyright (C) 2021  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from typing import Any, Iterable

from swh.core.api import remote_api_endpoint


class CountersInterface:
    @remote_api_endpoint("check")
    def check(self):
        """Dedicated method to execute some specific check per implementation.
        """
        ...

    @remote_api_endpoint("add")
    def add(self, collection: str, keys: Iterable[Any]) -> None:
        """Add the provided keys to the collection
           Only count new keys.
        """
        ...

    @remote_api_endpoint("count")
    def get_count(self, collection: str) -> int:
        """Return the number of keys for the provided collection"""
        ...



# Copyright (C) 2021  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import copy
import tempfile

from click.testing import CliRunner
from confluent_kafka import Producer
import pytest
import yaml

from swh.counters.cli import counters_cli_group
from swh.journal.serializers import value_to_kafka

CLI_CONFIG = """
counters:
    cls: redis
    host: %(redis_host)s
"""

JOURNAL_OBJECTS_CONFIG_TEMPLATE = """
journal:
    brokers:
        - {broker}
    prefix: {prefix}
    group_id: {group_id}
"""


def invoke(catch_exceptions, args, config="", *, redis_host):
    runner = CliRunner()
    with tempfile.NamedTemporaryFile("a", suffix=".yml") as config_fd:
        config_fd.write((CLI_CONFIG + config) % {"redis_host": redis_host})
        config_fd.seek(0)
        result = runner.invoke(counters_cli_group, ["-C" + config_fd.name] + args)
    if not catch_exceptions and result.exception:
        print(result.output)
        raise result.exception
    return result


def test__journal_client__worker_function_invoked(
    mocker, kafka_server, kafka_prefix, journal_config
):
    mock = mocker.patch("swh.counters.journal_client.process_journal_messages")
    mock.return_value = 1

    producer = Producer(
        {
            "bootstrap.servers": kafka_server,
            "client.id": "test-producer",
            "acks": "all",
        }
    )
    topic = f"{kafka_prefix}.content"
    value = value_to_kafka({"key": "value"})
    producer.produce(topic=topic, key=b"message1", value=value)

    invoke(
        False,
        # Missing --object-types (and no config key) will make the cli raise
        ["journal-client", "--stop-after-objects", "1", "--object-type", "content"],
        journal_config,
        redis_host="localhost",
    )

    assert mock.call_count == 1


def test__journal_client__missing_main_journal_config_key():
    """Missing configuration on journal should raise"""
    with pytest.raises(KeyError, match="journal"):
        invoke(
            catch_exceptions=False,
            args=["journal-client", "--stop-after-objects", "1",],
            config="",  # missing config will make it raise
            redis_host=None,
        )


def test__journal_client__missing_journal_config_keys():
    """Missing configuration on mandatory journal keys should raise"""
    kafka_prefix = "swh.journal.objects"
    journal_objects_config = JOURNAL_OBJECTS_CONFIG_TEMPLATE.format(
        broker="192.0.2.1", prefix=kafka_prefix, group_id="test-consumer"
    )
    journal_config = yaml.safe_load(journal_objects_config)

    for key in journal_config["journal"].keys():
        if key == "prefix":  # optional
            continue
        cfg = copy.deepcopy(journal_config)
        del cfg["journal"][key]  # make config incomplete
        yaml_cfg = yaml.dump(cfg)

        with pytest.raises(TypeError, match=f"{key}"):
            invoke(
                catch_exceptions=False,
                args=[
                    "journal-client",
                    "--stop-after-objects",
                    "1",
                    "--prefix",
                    kafka_prefix,
                    "--object-type",
                    "content",
                ],
                config=yaml_cfg,  # incomplete config will make the cli raise
                redis_host=None,
            )


def test__journal_client__missing_prefix_config_key(kafka_server):
    """Missing configuration on mandatory prefix key should raise"""

    journal_cfg_template = """
journal:
    brokers:
        - {broker}
    group_id: {group_id}
    """

    journal_cfg = journal_cfg_template.format(
        broker=kafka_server, group_id="test-consumer"
    )

    with pytest.raises(ValueError, match="prefix"):
        invoke(
            False,
            # Missing --prefix (and no config key) will make the cli raise
            [
                "journal-client",
                "--stop-after-objects",
                "1",
                "--object-type",
                "content",
            ],
            journal_cfg,
            redis_host=None,
        )


def test__journal_client__missing_object_types_config_key(kafka_server):
    """Missing configuration on mandatory object-types key should raise"""

    journal_cfg = JOURNAL_OBJECTS_CONFIG_TEMPLATE.format(
        broker=kafka_server, prefix="swh.journal.objects", group_id="test-consumer"
    )

    with pytest.raises(ValueError, match="object_types"):
        invoke(
            False,
            # Missing --object-types (and no config key) will make the cli raise
            ["journal-client", "--stop-after-objects", "1",],
            journal_cfg,
            redis_host=None,
        )


def test__journal_client__key_received(mocker, kafka_server):
    mock = mocker.patch("swh.counters.journal_client.process_journal_messages")
    mock.return_value = 1

    prefix = "swh.journal.objects"
    object_type = "content"
    topic = prefix + "." + object_type

    producer = Producer(
        {"bootstrap.servers": kafka_server, "client.id": "testproducer", "acks": "all",}
    )

    value = value_to_kafka({"key": "value"})
    key = b"message key"

    # Ensure empty messages are ignored
    producer.produce(topic=topic, key=b"emptymessage", value=None)
    producer.produce(topic=topic, key=key, value=value)
    producer.flush()

    journal_cfg = JOURNAL_OBJECTS_CONFIG_TEMPLATE.format(
        broker=kafka_server, prefix=prefix, group_id="test-consumer"
    )

    result = invoke(
        False,
        [
            "journal-client",
            "--stop-after-objects",
            "1",
            "--object-type",
            object_type,
            "--prefix",
            prefix,
        ],
        journal_cfg,
        redis_host=None,
    )

    # Check the output
    expected_output = "Processed 1 messages.\nDone.\n"
    assert result.exit_code == 0, result.output
    assert result.output == expected_output
    assert mock.call_args[0][0]["content"]
    assert len(mock.call_args[0][0]) == 1
    assert object_type in mock.call_args[0][0].keys()
    assert key == mock.call_args[0][0][object_type][0]



# Copyright (C) 2021  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information
import logging
import os
from typing import Any, Dict

from swh.core import config
from swh.core.api import RPCServerApp
from swh.counters import get_counters
from swh.counters.interface import CountersInterface

logger = logging.getLogger(__name__)

app = None


def make_app(config: Dict[str, Any]) -> RPCServerApp:
    """Initialize the remote api application.

    """
    app = RPCServerApp(
        __name__,
        backend_class=CountersInterface,
        backend_factory=lambda: get_counters(**config["counters"]),
    )

    handler = logging.StreamHandler()
    app.logger.addHandler(handler)

    app.add_url_rule("/", "index", index)

    return app


def index():
    return "SWH Counters API server"


def load_and_check_config(config_file: str) -> Dict[str, Any]:
    """Check the minimal configuration is set to run the api or raise an
       error explanation.

    Args:
        config_file: Path to the configuration file to load
        type: configuration type. For 'local' type, more
                    checks are done.

    Raises:
        Error if the setup is not as expected

    Returns:
        configuration as a dict

    """
    if not config_file:
        raise EnvironmentError("Configuration file must be defined")

    if not os.path.exists(config_file):
        raise FileNotFoundError("Configuration file %s does not exist" % (config_file,))

    cfg = config.read(config_file)
    if "counters" not in cfg:
        raise KeyError("Missing 'counters' configuration")

    return cfg


def make_app_from_configfile():
    """Run the WSGI app from the webserver, loading the configuration from
       a configuration file.

       SWH_CONFIG_FILENAME environment variable defines the
       configuration path to load.

    """
    global app

    if app is None:
        config_file = os.environ.get("SWH_CONFIG_FILENAME")
        api_cfg = load_and_check_config(config_file)

        app = make_app(api_cfg)

    return app



# Copyright (C) 2021  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from swh.counters.journal_client import process_journal_messages
from swh.counters.redis import Redis


def test__journal_client__all_keys(mocker):

    mock = mocker.patch("swh.counters.redis.Redis.add")

    redis = Redis(host="localhost")

    keys = {"coll1": [b"key1", b"key2"], "coll2": [b"key3", b"key4", b"key5"]}

    process_journal_messages(messages=keys, counters=redis)

    assert mock.call_count == 2

    first_call_args = mock.call_args_list[0]
    assert first_call_args[0][0] == "coll1"
    assert first_call_args[0][1] == keys["coll1"]

    second_call_args = mock.call_args_list[1]
    assert second_call_args[0][0] == "coll2"
    assert second_call_args[0][1] == keys["coll2"]
